import logging
import logging.config
from pathlib import Path
from typing import Any, Dict, Tuple

import pandas as pd
from jinja2 import Environment, PackageLoader
from vertica_python.vertica.cursor import Cursor

from dbdiff.vertica import (get_column_info_lookup,
                            implicit_dtype_comparison)

JINJA_ENV = Environment(loader=PackageLoader('dbdiff', 'templates'))
LOGGER = logging.getLogger(__name__)


def is_numeric_like(dtype: str):
    dtype_l = dtype.lower()
    return any({'int' in dtype_l, 'float' in dtype_l, 'numeric' in dtype_l})


def is_date_like(dtype: str):
    dtype_l = dtype.lower()
    return ('date' in dtype_l)


def check_primary_key(cur: Cursor,
                      schema: str, table: str,
                      join_cols: list) -> int:
    '''Given a list of columns return the # of records for which they are NOT
    a primary key.'''

    cur.execute(JINJA_ENV.get_template('table_rows.sql').render(schema_name=schema, table_name=table))
    r = cur.fetchall()
    n_rows = r[0]['COUNT']
    cur.execute(JINJA_ENV.get_template('table_rows_uniq.sql').render(schema_name=schema, table_name=table, join_cols=', '.join(join_cols)))
    n_distinct_rows = cur.fetchall()[0]['COUNT']
    return n_rows - n_distinct_rows


def get_all_col_info(cur: Cursor, schema, x_table, y_schema, y_table, exclude_columns_set, save_column_summary, save_column_summary_format):
    LOGGER.info('Getting column info for both tables.')
    x_table_info_lookup = get_column_info_lookup(cur, schema, x_table)
    y_table_info_lookup = get_column_info_lookup(cur, y_schema, y_table)

    def comparable_(x, y) -> bool:
        # this doesn't capture the case where they both could be converted to float to be compared (two hop conversions):
        if (x is None) or (y is None):
            return False
        left_or_right = implicit_dtype_comparison(x, y) or implicit_dtype_comparison(y, x)
        return left_or_right

    all_keys = list(x_table_info_lookup.keys()) + list(y_table_info_lookup.keys())
    all_col_info = {col: {'x_dtype': x_table_info_lookup.get(col, None),
                          'y_dtype': y_table_info_lookup.get(col, None),
                          'comparable': comparable_(x_table_info_lookup.get(col, None), y_table_info_lookup.get(col, None)),
                          'exclude': (col in exclude_columns_set)
                          } for col in all_keys}
    LOGGER.debug(all_col_info)
    all_col_info_df = pd.DataFrame(all_col_info).transpose()

    if save_column_summary:
        if save_column_summary_format.lower() == 'csv':
            all_col_info_df.to_csv(Path(x_table + '_col_info.csv'))
        if save_column_summary_format.lower() == 'pickle':
            all_col_info_df.to_pickle(Path(x_table + '_col_info.pkl'))
    else:
        LOGGER.info("All column info:\n" + all_col_info_df.to_string())
    LOGGER.info("Missing columns in x:\n" + all_col_info_df.loc[all_col_info_df.x_dtype.isnull(), :].to_string())
    LOGGER.info("Missing columns in y:\n" + all_col_info_df.loc[all_col_info_df.y_dtype.isnull(), :].to_string())
    LOGGER.debug(all_col_info_df.comparable)
    LOGGER.debug(~all_col_info_df.comparable)
    LOGGER.info("These columns have incompatible dtypes, specifically neither of them can be implicitly converted to the other:\n" + all_col_info_df.loc[(~all_col_info_df.comparable).astype('bool'), :].to_string())

    return all_col_info_df


def select_distinct_rows(cur: Cursor,
                         schema: str, table: str,
                         join_cols: list,
                         use_temp_tables: bool = False) -> Tuple[str, str]:
    '''Select only the rows that are distinct on join_cols.
    *Instead of deleting the rows, we'll select those without duplicates into a
    new table, and return the name of that new table.
    Delete is inefficient, see: https://www.vertica.com/docs/9.2.x/HTML/Content/Authoring/AnalyzingData/Optimizations/PerformanceConsiderationsForDELETEAndUPDATEQueries.htm
    And: https://www.vertica.com/blog/another-way-to-de-duplicate-table-rows-quick-tip/
    '''

    cur.execute(JINJA_ENV.get_template('table_drop.sql').render(schema_name=schema, table_name=(table + '_dedup')))
    q = JINJA_ENV.get_template('create_dedup.sql').render(
        schema_name=schema, table_name=table,
        table_name_dedup=(table + '_dedup'),
        group_cols=', '.join(join_cols),
        join_cols=' AND '.join(['x.{0} <=> y.{0}'.format(col) for col in join_cols]),
        use_temp_table=use_temp_tables
    )
    if use_temp_tables:
        q = JINJA_ENV.get_template('create_temp_table.sql').render(table_name=(table + '_dedup'), query=q)
    cur.execute(q)
    cur.execute('COMMIT;')
    cur.execute(JINJA_ENV.get_template('table_drop.sql').render(schema_name=schema, table_name=(table + '_dup')))
    q = JINJA_ENV.get_template('create_dup.sql').render(
        schema_name=schema, table_name=table,
        table_name_dup=(table + '_dup'),
        group_cols=', '.join(join_cols),
        join_cols=' AND '.join(['x.{0} <=> y.{0}'.format(col) for col in join_cols]),
        use_temp_table=use_temp_tables
    )
    if use_temp_tables:
        q = JINJA_ENV.get_template('create_temp_table.sql').render(table_name=(table + '_dup'), query=q)
    cur.execute(q)
    cur.execute('COMMIT;')

    return (schema, 'v_temp_schema')[use_temp_tables], '{table}_dedup'.format(table=table)


def create_joined_table(cur: Cursor, create_insert=False, **kwargs):
    """
    Joins two tables x and y.
    :param cur: vertica python Cursor
    :return: list - all queries run.
    """
    drop_q = JINJA_ENV.get_template('table_drop.sql').render(
        schema_name=kwargs['joined_schema'],
        table_name=kwargs['joined_table'])
    LOGGER.info(drop_q)
    cur.execute(drop_q)

    if create_insert:
        # these separately do CREATE TABLE and then
        # INSERT INTO
        create_q = JINJA_ENV.get_template('create_joined_table.sql').render(kwargs)
        LOGGER.info(create_q)
        cur.execute(create_q)
        insert_q = JINJA_ENV.get_template('insert_joined_table.sql').render(kwargs)
        LOGGER.info(insert_q)
        cur.execute(insert_q)
    else:
        # this does a SELECT INTO
        join_q = JINJA_ENV.get_template('create_joined_table_from_selectinto.sql').render(kwargs)
        LOGGER.info(join_q)
        cur.execute(join_q)

    LOGGER.info('COMMIT;')
    cur.execute('COMMIT;')

    table_rows_q = JINJA_ENV.get_template('table_rows.sql').render(
        schema_name=kwargs['joined_schema'],
        table_name=kwargs['joined_table'])
    LOGGER.info(table_rows_q)
    cur.execute(table_rows_q)
    r = cur.fetchall()
    joined_row_count = r[0]['COUNT']
    return joined_row_count


def get_unmatched_rows_straight(cur: Cursor,
                                x_schema: str, y_schema: str,
                                x_table: str, y_table: str,
                                join_cols: list,
                                max_rows_column: int) -> Dict[str, Dict[str, Any]]:
    all_keys_count = JINJA_ENV.get_template('all_keys_count.sql')
    all_keys_sample = JINJA_ENV.get_template('all_keys_sample.sql')

    results = {'left': {'count': 0, 'query': 'select ...', 'sample': pd.DataFrame()},
               'right': {'count': 0, 'query': 'select ...', 'sample': pd.DataFrame()}}

    for side in {'left', 'right'}:
        q = all_keys_count.render({'x_schema': x_schema, 'y_schema': y_schema,
                                   'x_table': x_table, 'y_table': y_table,
                                   'join_cols': join_cols, 'left': (side == 'left')})
        cur.execute(q)
        r = cur.fetchall()
        results[side]['count'] = r[0]['COUNT']
        results[side]['query'] = all_keys_sample.render({'x_schema': x_schema, 'y_schema': y_schema,
                                                         'x_table': x_table, 'y_table': y_table,
                                                         'join_cols': join_cols, 'max_rows_column': max_rows_column,
                                                         'left': (side == 'left')})
        cur.execute(results[side]['query'])
        results[side]['sample'] = pd.DataFrame(cur.fetchall())

    return results


def get_unmatched_rows(cur: Cursor,
                       x_schema: str, y_schema: str,
                       x_table: str, y_table: str,
                       join_cols: list,
                       max_rows_column: int) -> Dict[Any, Dict[str, Dict[str, Any]]]:
    '''
    Pull out rows that are unmatched between the two tables on the join columns.
    If looking at this hierarchically, we consider the join by
    key a, then key a+b (where a matched), then key a+b+c (where a+b matched), etc
    to see at what level we're missing things.
    '''
    results = {col: {'left': {'count': 0, 'query': 'select ...', 'sample': pd.DataFrame()},
                     'right': {'count': 0, 'query': 'select ...', 'sample': pd.DataFrame()}} for col in join_cols}

    first_key_count = JINJA_ENV.get_template('first_key_count.sql')
    first_key_t = JINJA_ENV.get_template('first_key_sample.sql')

    LOGGER.info('Getting rows that did not match on only the first join column: ' + join_cols[0] + '.')
    for side in {'left', 'right'}:
        q = first_key_count.render({'x_schema': x_schema, 'y_schema': y_schema,
                                    'x_table': x_table, 'y_table': y_table,
                                    'col': join_cols[0], 'left': (side == 'left')})
        cur.execute(q)
        r = cur.fetchall()
        results[join_cols[0]][side]['count'] = r[0]['COUNT']
        results[join_cols[0]][side]['query'] = first_key_t.render({'x_schema': x_schema, 'y_schema': y_schema,
                                                                   'x_table': x_table, 'y_table': y_table,
                                                                   'col': join_cols[0], 'max_rows': max_rows_column,
                                                                   'left': (side == 'left')})
        cur.execute(results[join_cols[0]][side]['query'])
        results[join_cols[0]][side]['sample'] = pd.DataFrame(cur.fetchall())

    sub_keys_count = JINJA_ENV.get_template('sub_keys_count.sql')
    sub_keys_t = JINJA_ENV.get_template('sub_keys_sample.sql')
    sub_keys_g = JINJA_ENV.get_template('sub_keys_grouped.sql')

    for i in range(1, len(join_cols)):
        LOGGER.info('Getting rows that did not match on the ' + str(i + 1) + '-nd/rd/th join column: ' + join_cols[i] + '.')
        LOGGER.info('This is equivalent to joining the tables on unique rows of ' + ','.join(join_cols[:(i + 1)]) + ' where all but the last already exist.')

        for side in {'left', 'right'}:
            q = sub_keys_count.render({'x_schema': x_schema, 'y_schema': y_schema,
                                       'x_table': x_table, 'y_table': y_table,
                                       'columns': join_cols[:(i + 1)], 'left': (side == 'left')})
            cur.execute(q)
            r = cur.fetchall()
            results[join_cols[i]][side]['count'] = r[0]['COUNT']
            results[join_cols[i]][side]['query'] = sub_keys_t.render({'x_schema': x_schema, 'y_schema': y_schema,
                                                                      'x_table': x_table, 'y_table': y_table,
                                                                      'columns': join_cols[:(i + 1)], 'max_rows': max_rows_column,
                                                                      'left': (side == 'left')})
            cur.execute(results[join_cols[i]][side]['query'])
            results[join_cols[i]][side]['sample'] = pd.DataFrame(cur.fetchall())
            results[join_cols[i]][side]['query_grouped'] = sub_keys_g.render({'x_schema': x_schema, 'y_schema': y_schema,
                                                                              'x_table': x_table, 'y_table': y_table,
                                                                              'columns': join_cols[:(i + 1)], 'max_rows': max_rows_column,
                                                                              'left': (side == 'left')})
            cur.execute(results[join_cols[i]][side]['query_grouped'])
            results[join_cols[i]][side]['sample_grouped'] = pd.DataFrame(cur.fetchall())

    return results


def create_diff_table(cur: Cursor,
                      schema: str, table: str,
                      join_cols: list, all_col_info_df: pd.DataFrame) -> str:
    drop_q = JINJA_ENV.get_template('table_drop.sql').render(schema_name=schema, table_name=table)
    # so simple that putting into a template would make this harder to follow...
    q = 'CREATE TABLE {schema}.{table} ( {columns}, column_name VARCHAR(255) );'.format(
        schema=schema,
        table=table,
        columns=', '.join(all_col_info_df.loc[all_col_info_df.index.isin(join_cols)].apply(lambda x: ' '.join([x.name, x.x_dtype]), axis=1).values)
    )
    cur.execute(drop_q)
    cur.execute(q)
    return q


def insert_diff_table(cur: Cursor, **kwargs) -> None:
    cur.execute(JINJA_ENV.get_template('insert_diff.sql').render(kwargs))
    cur.execute('COMMIT;')


def get_diff_rows(cur: Cursor,
                  output_schema: str,
                  x_table: str,
                  join_cols: list,
                  max_rows_all: int,
                  skip_row_total: bool = False) -> dict:
    LOGGER.debug("Getting diff rows")
    # first get the count
    q = JINJA_ENV.get_template('table_rows.sql').render(
        schema_name=output_schema,
        table_name=(x_table + '_DIFF'))
    LOGGER.info(q)
    cur.execute(q)
    diff_total_count = cur.fetchall()[0]['COUNT']
    if skip_row_total:
        LOGGER.debug("Skipping sample of rows with differences, query to get that sample, and the total # of rows with > 0 differences. Returning only 'total_count', the sum of cell-by-cell differences.")
        return {'total_count': diff_total_count}

    q = JINJA_ENV.get_template('table_rows_uniq.sql').render(schema_name=output_schema, table_name=(x_table + '_DIFF'), join_cols=', '.join(join_cols))
    LOGGER.info(q)
    cur.execute(q)
    diff_row_count = cur.fetchall()[0]['COUNT']

    # we'll pull all columns from the joined table
    q = JINJA_ENV.get_template('diff_rows_sample.sql').render(
        schema_name=output_schema,
        joined_table=(x_table + '_JOINED'),
        diff_table=(x_table + '_DIFF'),
        group_cols=', '.join(join_cols),
        join_cols=' AND '.join(['x.{0} <=> joined.{0}'.format(col) for col in join_cols])
    )
    LOGGER.info(q)
    cur.execute(q + ' LIMIT ' + str(max_rows_all))
    diff_rows = pd.DataFrame(cur.fetchall())

    return {'query': q, 'sample': diff_rows,
            'count': diff_row_count, 'total_count': diff_total_count}


def get_diff_rows_from_joined(cur: Cursor,
                              grouped_column_diffs: dict,
                              output_schema: str,
                              x_table: str,
                              join_cols: list,
                              max_rows_all: int,
                              skip_row_total: bool = False) -> dict:
    '''Get diff rows from joined table.

    Non self-explanatory argument specifics:

    - grouped_column_diffs:
    - max_rows_all: number of rows to get for the sample (only relevant if skip_row_total=F)
    - skip_row_total: skip sample of rows with differences, query to get that sample, and the total # of rows with > 0 differences. Return only 'total_count', the sum of cell-by-cell differences.

    Returned data specifics:

    - dict with 4 keys:
        - total_count: total number of cell-by-cell differences between the two tables.
        - query: query to get a sample of rows with >0 differences.
        - sample: dataframe of those sample rows
        - count: count of rows with >0 differences.
    '''
    LOGGER.debug("Getting diff rows: get_diff_rows_from_joined()")

    diff_total_count = sum([info['count'] for info in grouped_column_diffs.values()])
    if skip_row_total:
        LOGGER.debug("Skipping sample of rows with differences, query to get that sample, and the total # of rows with > 0 differences. Returning only 'total_count', the sum of cell-by-cell differences.")
        return {'total_count': diff_total_count}

    LOGGER.info(grouped_column_diffs)
    q = JINJA_ENV.get_template('joined_rows_count.sql').render(
        joined_schema=output_schema,
        joined_table=(x_table + '_JOINED'),
        columns=grouped_column_diffs.keys()
    )
    LOGGER.info(q)
    cur.execute(q)
    diff_row_count = cur.fetchall()[0]['COUNT']

    # we'll pull all columns from the joined table
    q = JINJA_ENV.get_template('joined_rows_sample.sql').render(
        joined_schema=output_schema,
        joined_table=(x_table + '_JOINED'),
        columns=grouped_column_diffs.keys()
    )
    LOGGER.info(q)
    cur.execute(q + ' LIMIT ' + str(max_rows_all))
    diff_rows = pd.DataFrame(cur.fetchall())

    return {'query': q, 'sample': diff_rows,
            'count': diff_row_count, 'total_count': diff_total_count}


def get_diff_columns(cur: Cursor, output_schema: str, x_table: str) -> pd.DataFrame:
    LOGGER.debug("Getting diff columns")
    # The # of columns has a hard limit (~1600 in Vertica?) so don't worry about
    # pulling the count first or limiting the results
    q = JINJA_ENV.get_template('diff_column_summary.sql').render(
        schema_name=output_schema,
        table_name=(x_table + '_DIFF'))
    cur.execute(q)
    return pd.DataFrame(cur.fetchall())


def get_column_diffs(diff_columns: pd.DataFrame, cur: Cursor,
                     output_schema: str,
                     x_schema: str, x_table: str,
                     y_schema: str, y_table: str,
                     join_cols: list,
                     max_rows_column: int,
                     all_col_info_df: pd.DataFrame,
                     hierarchical: bool = False) -> dict:
    LOGGER.debug("Getting column diffs")
    # get total count, list of most common differing pairs for each column
    # list of (count, query, df)
    grouped_column_diffs = {row.column_name: {'count': row['COUNT']} for i, row in diff_columns.iterrows()}

    for column_name, info in grouped_column_diffs.items():
        LOGGER.info('Getting detailed diff for column: ' + str(column_name) + ' with ' + str(info['count']) + ' differences.')
        q = JINJA_ENV.get_template('diff_column.sql').render(
            column=column_name,
            joined_schema=output_schema, joined_table=(x_table + '_JOINED'),
            diff_schema=output_schema, diff_table=(x_table + '_DIFF'),
            group_cols=', '.join(join_cols),
            join_cols=' AND '.join(['diff.{0} <=> joined.{0}'.format(col) for col in join_cols]),
        )
        info['q'] = q
        q_raw = JINJA_ENV.get_template('diff_column_raw.sql').render(
            column=column_name,
            joined_schema=output_schema, joined_table=(x_table + '_JOINED'),
            diff_schema=output_schema, diff_table=(x_table + '_DIFF'),
            join_cols=join_cols,
            join_cols_join=' AND '.join(['diff.{0} <=> joined.{0}'.format(col) for col in join_cols]),
        )
        info['q_raw'] = q_raw
        cur.execute(q + ' LIMIT ' + str(max_rows_column))
        info['df'] = pd.DataFrame(cur.fetchall())
        cur.execute(q_raw + ' LIMIT ' + str(max_rows_column))
        info['df_raw'] = pd.DataFrame(cur.fetchall())
        if hierarchical:
            for schema, table, side in ((x_schema, x_table, 'x'), (y_schema, y_table, 'y')):
                for limit in (None, max_rows_column):
                    q_h = JINJA_ENV.get_template('diff_column_hier.sql').render(
                        column=column_name,
                        diff_schema=output_schema,
                        diff_table=(x_table + '_DIFF'),
                        join_cols=', '.join(join_cols),
                        first_join_col=join_cols[0],
                        schema=schema,
                        table=table,
                        limit=limit
                    )
                    if limit is None:
                        info['q_h_' + side] = q_h
                    else:
                        cur.execute(q_h)
                        info['df_h_' + side] = pd.DataFrame(cur.fetchall())
        row = all_col_info_df.loc[column_name, :]
        is_numeric = (is_numeric_like(row.x_dtype) and is_numeric_like(row.y_dtype))
        is_date = (is_date_like(row.x_dtype) and is_date_like(row.y_dtype))
        if is_numeric or is_date:
            info['q_n'] = JINJA_ENV.get_template('diff_column_numeric_diffs_binned.sql').render(
                column=column_name,
                joined_schema=output_schema, joined_table=(x_table + '_JOINED'),
                diff_schema=output_schema, diff_table=(x_table + '_DIFF'),
                group_cols=', '.join(join_cols),
                join_cols=' AND '.join(['diff.{0} <=> joined.{0}'.format(col) for col in join_cols]),
                tiles=min({max({1, info['count']}), 10}))
            cur.execute(info['q_n'])
            info['df_n'] = pd.DataFrame(cur.fetchall())
            info['q_n_sample'] = JINJA_ENV.get_template('diff_column_numeric_diffs_sorted.sql').render(
                column=column_name,
                joined_schema=output_schema, joined_table=(x_table + '_JOINED'),
                diff_schema=output_schema, diff_table=(x_table + '_DIFF'),
                join_cols=join_cols,
                join_cols_join=' AND '.join(['diff.{0} <=> joined.{0}'.format(col) for col in join_cols]),
            )
            cur.execute(info['q_n_sample'] + ' LIMIT ' + str(max_rows_column))
            info['df_n_sample'] = pd.DataFrame(cur.fetchall())
    return grouped_column_diffs


def get_column_diffs_from_joined(cur: Cursor,
                                 output_schema: str,
                                 x_schema: str, x_table: str,
                                 y_schema: str, y_table: str,
                                 join_cols: list,
                                 max_rows_column: int,
                                 all_col_info_df: pd.DataFrame,
                                 comparable_filter,
                                 hierarchical: bool = False) -> dict:
    '''Get column-by-column diffs directly from the joined table.

    Non self-explanatory argument specifics:

    - max_rows_column: number of rows to pull for sample differing cells on each column.
    - all_col_info_df: dataframe with the following columns:
        - index: column names.
        - x_dtype.
        - y_dtype.
    - comparable_filter: an 0/1 index on all_col_info_df to filter on columns to compare.
        - in cli.py, this filter/index is set using datatype matching and the user-supplied list of columns to exclude.
    - hierarchical: if true, additional outputs are included for each columns that are samples with the join keys.

    Returned data specifics:
    - dict grouped_column_diffs:
        - each `key` is a string column name for columns that matches on name between x and y tables (and are comparable on dtype, not excluded by user-supplied list).
        - each `value` is a dictionary with the following keys:
            - {'count': diff_count, 'df': df, 'df_raw': df_raw, 'q': q, 'q_raw': q_raw}.
            - if `hierarchical` is true: `{q,d}_h_{x,y}` (q for query, d for dataframe sample) from x and y tables, respectively.
            - if numeric of date: `{q,df}_n{,_sample}` (q for query, df for dataframe), the `_sample` is the biggest diffs, while the former are the binned differences.
    '''
    column_list_to_compare = all_col_info_df.loc[comparable_filter & ~all_col_info_df.index.isin(join_cols), :].index.values
    LOGGER.info("Getting column diffs for columns:")
    LOGGER.info(",".join(column_list_to_compare))
    grouped_column_diffs = {}

    for column in column_list_to_compare:
        LOGGER.info("=" * 80)
        LOGGER.info(column)
        joined_count_q = JINJA_ENV.get_template('joined_count.sql').render(
            column=column,
            joined_schema=output_schema,
            joined_table=(x_table + '_JOINED')
        )
        LOGGER.info(joined_count_q)
        cur.execute(joined_count_q)
        diff_count = cur.fetchall()[0]['COUNT']
        if diff_count > 0:
            LOGGER.info('Getting detailed diff for column: ' + str(column) + ' with ' + str(diff_count) + ' differences.')
            q = JINJA_ENV.get_template('joined_column.sql').render(
                column=column,
                joined_schema=output_schema, joined_table=(x_table + '_JOINED')
            )
            q_raw = JINJA_ENV.get_template('joined_column_raw.sql').render(
                column=column,
                joined_schema=output_schema, joined_table=(x_table + '_JOINED'),
                join_cols=join_cols
            )
            LOGGER.info(q)
            cur.execute(q + ' LIMIT ' + str(max_rows_column))
            df = pd.DataFrame(cur.fetchall())
            LOGGER.info(q_raw)
            cur.execute(q_raw + ' LIMIT ' + str(max_rows_column))
            df_raw = pd.DataFrame(cur.fetchall())
            grouped_column_diffs[column] = {'count': diff_count, 'df': df, 'df_raw': df_raw, 'q': q, 'q_raw': q_raw}
            LOGGER.info(grouped_column_diffs[column])

            if hierarchical:
                for schema, table, side in ((x_schema, x_table, 'x'), (y_schema, y_table, 'y')):
                    for limit in (None, max_rows_column):
                        q_h = JINJA_ENV.get_template('joined_column_hier.sql').render(
                            column=column,
                            joined_schema=output_schema, joined_table=(x_table + '_JOINED'),
                            join_cols=join_cols,
                            schema=schema,
                            table=table,
                            limit=limit
                        )
                        if limit is None:
                            grouped_column_diffs[column]['q_h_' + side] = q_h
                        else:
                            cur.execute(q_h)
                            grouped_column_diffs[column]['df_h_' + side] = pd.DataFrame(cur.fetchall())
            row = all_col_info_df.loc[column, :]
            is_numeric = (is_numeric_like(row.x_dtype) and is_numeric_like(row.y_dtype))
            is_date = (is_date_like(row.x_dtype) and is_date_like(row.y_dtype))
            if is_numeric or is_date:
                grouped_column_diffs[column]['q_n'] = JINJA_ENV.get_template('joined_column_numeric_diffs_binned.sql').render(
                    column=column,
                    joined_schema=output_schema, joined_table=(x_table + '_JOINED'),
                    tiles=min({max({1, grouped_column_diffs[column]['count']}), 10}))
                cur.execute(grouped_column_diffs[column]['q_n'])
                grouped_column_diffs[column]['df_n'] = pd.DataFrame(cur.fetchall())
                grouped_column_diffs[column]['q_n_sample'] = JINJA_ENV.get_template('joined_column_numeric_diffs_sorted.sql').render(
                    column=column,
                    joined_schema=output_schema, joined_table=(x_table + '_JOINED'),
                    join_cols=join_cols
                )
                cur.execute(grouped_column_diffs[column]['q_n_sample'] + ' LIMIT ' + str(max_rows_column))
                grouped_column_diffs[column]['df_n_sample'] = pd.DataFrame(cur.fetchall())
        else:
            LOGGER.info('NOT getting detailed diff for column: ' + str(column) + ' with ' + str(diff_count) + ' differences.')
    LOGGER.info(len(grouped_column_diffs))
    grouped_column_diffs_sorted = {x: grouped_column_diffs[x] for x in sorted(grouped_column_diffs.keys(), key=lambda x: grouped_column_diffs[x]['count'], reverse=True)}
    LOGGER.info(len(grouped_column_diffs_sorted))
    return grouped_column_diffs_sorted



import logging
import os
import ssl
from contextlib import contextmanager

import pandas as pd
import requests
import vertica_python
from dotenv import find_dotenv, load_dotenv
from jinja2 import Environment, PackageLoader
from vertica_python.vertica.cursor import Cursor

JINJA_ENV = Environment(loader=PackageLoader('dbdiff', 'templates'))
LOGGER = logging.getLogger(__name__)


def get_column_info(cur: Cursor, schema_name: str,
                    table_name: str) -> pd.DataFrame:
    column_template = JINJA_ENV.get_template('table_columns.sql')
    cur.execute(column_template.render(schema_name=schema_name,
                                       table_name=table_name))
    # go ahead and convert (before checking size), overhead is low.
    df = pd.DataFrame(cur.fetchall())
    if df.shape[0] == 0:
        raise RuntimeError('{schema}.{table} has no columns.'.format(schema=schema_name, table=table_name))
    return df


def get_column_info_lookup(cur: Cursor, schema_name: str,
                           table_name: str) -> dict:
    x_table_info = get_column_info(cur, schema_name, table_name)
    LOGGER.info("Column info Vertica for " + schema_name + "." + table_name + ":\n" + x_table_info.head().to_string() + "\n...")
    x_table_info['column_name'] = x_table_info['column_name'].str.lower()
    return {r.column_name: r.data_type for i, r in x_table_info.iterrows()}


def get_table_exists(cur: Cursor, schema_name: str,
                     table_name: str) -> bool:
    table_exists_template = JINJA_ENV.get_template('table_exists.sql')
    cur.execute(table_exists_template.render(schema_name=schema_name,
                                             table_name=table_name))
    return (cur.fetchall()[0]['COUNT'] == 1)


def implicit_dtype_comparison(x_dtype: str, y_dtype: str) -> bool:
    '''Can x_dtype be implicitly converted to y_dtype?
    https://www.vertica.com/docs/9.2.x/HTML/Content/Authoring/SQLReferenceManual/DataTypes/DataTypeCoercionChart.htm
    '''
    source = x_dtype.upper()
    target = y_dtype.upper()
    if 'INT' in source:  # {'INT', 'INTEGER'} since numeric can look like NUMERIC(18,2), etc
        return ('BOOL' in target or
                'NUMERIC' in target or
                'FLOAT' in target or
                'INT' in target)
    elif 'NUMERIC' in source:
        return ('FLOAT' in target or
                'NUMERIC' in target)
    # don't worry about LONG VARCHAR not converting to VARCHAR right now
    elif 'CHAR' in source:
        return ('CHAR' in target or
                'FLOAT' in target)
    elif 'DATE' in source and 'DATE' in target:
        return True
    else:
        return source == target


@contextmanager
def get_cur() -> Cursor:
    '''Build a connection.

    For connection options,
    this function will look for a file, recursively up from this directory,
    named .config.sh.
    If such a file exists, the keys VERTICA_*
    will be used to make the connection.
    Else, it will use the available environment variables.

    For SSL, if the environment variable CERT_LINK is set,
    this function will get the file at that uri and use it.
    If no CERT_LINK is set and VERTICA_SSL is set one of:
    ['1', 'true', 'yes', 'please']
    then this function will use the system's default context
    (openssl's context (set this with openssl's env variable config)).
    If neither CERT_LINK nor VERTICA_SSL are set,
    this will not use SSL.
    '''
    load_dotenv(find_dotenv('.config.sh'))

    conninfo = dict(host=os.environ.get('VERTICA_HOST', '').strip(),
                    port=int(os.environ.get('VERTICA_PORT', '').strip()),
                    database=os.environ.get('VERTICA_DATABASE', '').strip(),
                    user=os.environ.get('VERTICA_USERNAME', '').strip(),
                    password=os.environ.get('VERTICA_PASSWORD', '').strip(),
                    connection_timeout=int(os.environ.get(
                        'VERTICA_CONNECTION_TIMEOUT', '36000').strip()),
                    # read_timeout=int(os.environ.get(
                    #     'VERTICA_READ_TIMEOUT', '36000').strip()),
                    # this is in the docs, but not used!
                    unicode_error=os.environ.get(
                        'VERTICA_UNICODE_ERROR', 'strict').strip())
    # let this one use the vertica-python default:
    # connection_load_balance=True
    LOGGER.debug(conninfo)

    if os.environ.get('CERT_LINK') is not None:
        LOGGER.debug('getting cert from uri')
        resp = requests.get(str(os.environ.get('CERT_LINK')))
        resp.raise_for_status()
        ssl_data = resp.content

        context = ssl.create_default_context(cadata=ssl_data.decode('ascii', 'ignore'))
        context.check_hostname = False
        conninfo['ssl'] = context
    else:
        use_ssl_env = os.environ.get('VERTICA_SSL')
        if (use_ssl_env is not None) and ((use_ssl_env == '1') or (use_ssl_env.lower() in {'true', 'yes', 'please'})):
            context = ssl.create_default_context()
            context.check_hostname = False
            conninfo['ssl'] = context

    with vertica_python.connect(**conninfo) as conn:
        with conn.cursor('dict') as cur:
            try:
                yield cur
            finally:
                conn.close()



import json
import logging
import logging.config
from pathlib import Path

import click
import pandas as pd
from jinja2 import Environment, PackageLoader

from dbdiff.main import (check_primary_key, create_diff_table,
                         create_joined_table, get_all_col_info,
                         get_column_diffs, get_column_diffs_from_joined,
                         get_diff_columns, get_diff_rows,
                         get_diff_rows_from_joined, get_unmatched_rows,
                         get_unmatched_rows_straight, insert_diff_table,
                         select_distinct_rows)
from dbdiff.report import excel_report, html_report
from dbdiff.vertica import get_cur

JINJA_ENV = Environment(loader=PackageLoader('dbdiff', 'templates'))
DEFAULT_LOGGING_CONFIG = Path(__file__).with_name('logging.json')
LOGGER = logging.getLogger(__name__)


def initialize_logging(config: Path) -> None:
    """
    Initialize logging configuration from JSON config file.
    """
    with config.open() as fobj:
        dict_config = json.load(fobj)
    logging.config.dictConfig(dict_config)


@click.command()
@click.argument('schema')
@click.argument('x_table')
@click.argument('y_table')
@click.argument('join_cols')
@click.option('--y-schema', default=None, help='If the schema for the y_table is different, specify it.')
@click.option('--output-schema', default=None, help='If you want the schema for the output tables to be different, specify it.')
@click.option('--drop-output-tables', is_flag=True, help='Drop the joined and diff tables created and used here.')
@click.option('--x-table-query', is_flag=True, help="If X_TABLE is not a table in Vertica, but rather a query stored in a file, add this flag and the query will be read and instantiated into a temporary table. Ex: 'temp_xtable_name_to_use.sql'.")
@click.option('--y-table-query', is_flag=True, help='If Y_TABLE is not a table in Vertica, but rather a query stored in a file, add this flag and the query will be read and instantiated into a temporary table.')
@click.option('--exclude-columns', default="", help='Comma separated string of column names to exclude.')
@click.option('--hierarchical-join', is_flag=True, help='If multiple join keys, and join key #2 is a subset of join key #1. We expect matches for all of #1 from both tables even if we dont match on #1 and #2. This way, we can have more nuanced output by first breaking out missing on the first key.')
@click.option('--max-rows-all', default=10, help='Limit of full rows to pull that have differences.', show_default=True)
@click.option('--max-rows-column', default=10, help='Limit of grouped and raw column level differences to pull.', show_default=True)
@click.option('--output-format', type=click.Choice(['HTML', 'XLSX'], case_sensitive=False), default="HTML")
@click.option('--save-column-summary', is_flag=True, help='Save the column dtype and match summary.')
@click.option('--save-column-summary-format', type=click.Choice(['CSV', 'PICKLE'], case_sensitive=False), default="CSV")
@click.option('--skip-row-total', is_flag=True, help='Skip counting the total # of rows with differences, only use cell differences.')
@click.option('--use-diff-table', is_flag=True, help='Use a diff table in the middle.')
@click.option('--logging-config', type=Path, default=DEFAULT_LOGGING_CONFIG)
@click.option('--case-insensitive', is_flag=True, help='If using this flag, all case sensitivity is turned off.')
def main(schema: str, x_table: str, y_table: str,
         join_cols: str, y_schema: str, output_schema: str, drop_output_tables: bool,
         x_table_query: bool, y_table_query: bool, exclude_columns: str,
         hierarchical_join: bool, max_rows_all: int, max_rows_column: int,
         output_format: str, save_column_summary: bool,
         save_column_summary_format: str, skip_row_total: bool,
         use_diff_table: bool, logging_config: Path, case_insensitive: bool):
    """Compare two flat files X_TABLE and Y_TABLE, using Vertica as the join engine.
    Assume they are both in the same schema = SCHEMA.
    Join them on the columns in comma-separated string JOIN_COLS.
    Expects that the join columns have matching data type or will implicitly cast for comparison,
    and implicity cast into the type in [X_TABLE] for the JOINED table.
    Expects that all other columns with matchings names (those that will be compared)
    can be compared directly (it will cast all dtypes for comparison to the type in X_TABLE).

    Will drop [X_TABLE]_DIFF (if --use-diff-table) and will drop [X_TABLE]_JOINED if they exist."""
    # default y_schema to be the same as x
    if y_schema is None:
        y_schema = schema
    if output_schema is None:
        output_schema = schema
    join_cols_list = list(map(lambda x: x.lower(), join_cols.split(',')))
    exclude_columns_set = set(map(lambda x: x.lower(), exclude_columns.split(',')))
    initialize_logging(logging_config)

    with get_cur() as cur:
        if case_insensitive:
            LOGGER.info('Setting to case insensitive.')
            cur.execute("SET LOCALE TO 'en_US@colstrength=1';")
            # clear the results
            cur.fetchall()
        if x_table_query:
            with open(x_table, 'r') as f:
                q = f.read()
            x_table = Path(x_table).stem
            LOGGER.info('Creating temp table from query for x.')
            cur.execute(JINJA_ENV.get_template('create_temp_table.sql').render(table_name=x_table, query=q))
            schema = 'v_temp_schema'
        if y_table_query:
            with open(y_table, 'r') as f:
                q = f.read()
            y_table = Path(y_table).stem
            LOGGER.info('Creating temp table from query for y.')
            cur.execute(JINJA_ENV.get_template('create_temp_table.sql').render(table_name=y_table, query=q))
            y_schema = 'v_temp_schema'

        all_col_info_df = get_all_col_info(
            cur,
            schema,
            x_table,
            y_schema,
            y_table,
            exclude_columns_set,
            save_column_summary,
            save_column_summary_format
        )
        comparable_filter = (~all_col_info_df.exclude & all_col_info_df.comparable & ~all_col_info_df.x_dtype.isnull() & ~all_col_info_df.y_dtype.isnull())
        # check that the join cols exist on both tables
        for col in join_cols_list:
            if all_col_info_df.loc[comparable_filter & (all_col_info_df.index == col), :].shape[0] == 0:
                raise RuntimeError('Column `{0}` not in comparable columns (missing from one, both, or bad dtype). Here is the info we do have about that col:\n'.format(col) + all_col_info_df.loc[col, :].to_string())

        LOGGER.info('Checking primary keys.')
        x = check_primary_key(cur=cur, schema=schema, table=x_table, join_cols=join_cols_list)
        y = check_primary_key(cur=cur, schema=y_schema, table=y_table, join_cols=join_cols_list)
        dedup_info = {x_table: {'count': x}, y_table: {'count': y}}

        # hard stop on primary key:
        # assert x == 0, '# non distinct rows in ' + x_table + ' is ' + str(x)
        # assert y == 0, '# non distinct rows in ' + y_table + ' is ' + str(y)

        hierarchical_join_info = {}
        if hierarchical_join:
            LOGGER.info('Getting rows that are missing on each join key.')
            hierarchical_join_info = get_unmatched_rows(
                cur=cur,
                x_schema=schema,
                y_schema=y_schema,
                x_table=x_table,
                y_table=y_table,
                join_cols=join_cols_list,
                max_rows_column=max_rows_column
            )

        # create sub-tables to allow a comparison:
        if x != 0:
            LOGGER.info('X table was not unique on join keys, creating _dedup and _dup versions.')
            schema, x_table = select_distinct_rows(
                cur,
                schema,
                x_table,
                join_cols_list,
                use_temp_tables=(drop_output_tables or x_table_query)
            )
        if y != 0:
            LOGGER.info('Y table was not unique on join keys, creating _dedup and _dup versions.')
            y_schema, y_table = select_distinct_rows(
                cur,
                y_schema,
                y_table,
                join_cols_list,
                use_temp_tables=(drop_output_tables or y_table_query)
            )

        LOGGER.info('Getting rows that did not match (not in joined table) after deduping.')
        missing_join_info = get_unmatched_rows_straight(
            cur=cur,
            x_schema=schema,
            y_schema=y_schema,
            x_table=x_table,
            y_table=y_table,
            join_cols=join_cols_list,
            max_rows_column=max_rows_column
        )

        # build the joined table
        LOGGER.info('Building joined table ' + (x_table + '_JOINED'))
        joined_row_count = create_joined_table(
            cur=cur,
            x_schema=schema,
            y_schema=y_schema,
            x_table=x_table,
            y_table=y_table,
            join_cols=join_cols_list,
            compare_cols=all_col_info_df.loc[comparable_filter, :],
            joined_schema=output_schema,
            joined_table=(x_table + '_JOINED')
        )

        if use_diff_table:
            # build the diff table
            LOGGER.info('Building diff table ' + (x_table + '_DIFF.'))
            create_diff_table(
                cur=cur,
                schema=output_schema,
                table=(x_table + '_DIFF'),
                join_cols=join_cols_list,
                all_col_info_df=all_col_info_df
            )
            for column in all_col_info_df.loc[comparable_filter & ~all_col_info_df.index.isin(join_cols_list), :].index.values:
                LOGGER.info('Inserting column ' + column + ' into diff table.')
                insert_diff_table(
                    cur=cur,
                    joined_schema=output_schema,
                    joined_table=(x_table + '_JOINED'),
                    diff_schema=output_schema,
                    diff_table=(x_table + '_DIFF'),
                    join_cols=join_cols_list,
                    column=column
                )

            ############################################################################
            # Result 1: Get rows with at least N=1 difference (count, query, dataframe),
            ############################################################################
            diff_summary = get_diff_rows(cur, output_schema, x_table, join_cols_list, max_rows_all, skip_row_total)

            ############################################################################
            # Result 2: Get ordered list of columns by # of differences (query, dataframe).
            ############################################################################
            diff_columns = get_diff_columns(cur, output_schema, x_table)

            ############################################################################
            # Result 3: Get detailed column diffs.
            ############################################################################
            grouped_column_diffs = get_column_diffs(diff_columns, cur, output_schema, schema, x_table, y_schema, y_table, join_cols_list, max_rows_column, all_col_info_df, hierarchical_join)

        else:
            grouped_column_diffs = get_column_diffs_from_joined(
                cur=cur,
                output_schema=output_schema,
                x_schema=schema,
                x_table=x_table,
                y_schema=y_schema,
                y_table=y_table,
                join_cols=join_cols_list,
                max_rows_column=max_rows_column,
                all_col_info_df=all_col_info_df,
                comparable_filter=comparable_filter,
                hierarchical=hierarchical_join
            )
            diff_summary = get_diff_rows_from_joined(
                cur=cur,
                grouped_column_diffs=grouped_column_diffs,
                output_schema=output_schema,
                x_table=x_table,
                join_cols=join_cols_list,
                max_rows_all=max_rows_all,
                skip_row_total=skip_row_total
            )

        if output_format == 'HTML':
            report = html_report(
                x_schema=schema,
                y_schema=y_schema,
                x_table=x_table,
                y_table=y_table,
                join_cols=join_cols_list,
                total_row_count=joined_row_count,
                column_info=grouped_column_diffs,
                column_match_info=all_col_info_df,
                missing_join_info=missing_join_info,
                hierarchical_join_info=hierarchical_join_info,
                dedup_info=dedup_info,
                diff_summary=diff_summary
            )
            with open(x_table + '_report.html', 'w') as f:
                f.write(report)
        elif output_format == 'XLSX':
            reports = excel_report(x_schema=schema,
                                   y_schema=y_schema,
                                   x_table=x_table,
                                   y_table=y_table,
                                   join_cols=join_cols_list,
                                   total_row_count=joined_row_count,
                                   column_info=grouped_column_diffs,
                                   column_match_info=all_col_info_df,
                                   missing_join_info=missing_join_info,
                                   hierarchical_join_info=hierarchical_join_info,
                                   diff_summary=diff_summary)
            writer = pd.ExcelWriter(x_table + '_report.xlsx', engine='xlsxwriter')
            for sheet_name, df in reports:
                df.to_excel(writer, sheet_name=sheet_name, index=False)
            writer.save()

        if drop_output_tables:
            LOGGER.info("Dropping output tables. WARNING: queries in the report won't work!")
            cur.execute(JINJA_ENV.get_template('table_drop.sql').render(schema_name=output_schema, table_name=(x_table + '_JOINED')))
            if use_diff_table:
                cur.execute(JINJA_ENV.get_template('table_drop.sql').render(schema_name=output_schema, table_name=(x_table + '_DIFF')))



import pandas as pd
from jinja2 import Environment, PackageLoader

MAX_EXCEL_SHEET_NAME_LEN = 31
JINJA_ENV = Environment(loader=PackageLoader('dbdiff', 'templates'))


def html_report(x_schema: str, y_schema: str, x_table: str, y_table: str, join_cols: list,
                diff_summary: dict, total_row_count: int,
                column_info: dict,
                column_match_info: pd.DataFrame,
                missing_join_info: dict, hierarchical_join_info: dict,
                dedup_info: dict) -> str:

    def comma(value, format='{0:,d}'):
        return format.format(value)

    def code(value, codeclass='plaintext'):
        return '<code class="{0}">{1}</code>'.format(codeclass, value)

    def dfhtml(df, classes=["table", "table-bordered", "table-striped", "table-hover", "table-sm"]):
        return df.to_html(index=False, classes=classes)

    JINJA_ENV.filters['comma'] = comma
    JINJA_ENV.filters['code'] = code
    JINJA_ENV.filters['dfhtml'] = dfhtml

    t = JINJA_ENV.get_template('report.html')
    if len(list(column_info.values())) > 0:
        max_differences = list(column_info.values())[0]['count']
    else:
        max_differences = 0
    missing_join_info = {x_table: missing_join_info['right'], y_table: missing_join_info['left']}
    for col in hierarchical_join_info:
        hierarchical_join_info[col] = {x_table: hierarchical_join_info[col]['right'], y_table: hierarchical_join_info[col]['left']}
    column_match_info['uncomparable'] = (~column_match_info.comparable) & (~column_match_info.x_dtype.isnull()) & (~column_match_info.y_dtype.isnull())
    print(column_match_info['uncomparable'])
    return t.render({'x_schema': x_schema, 'y_schema': y_schema,
                     'x_table': x_table, 'y_table': y_table,
                     'join_cols': join_cols,
                     'diff_summary': diff_summary,
                     'total_row_count': total_row_count,
                     'column_info': column_info,
                     'max_differences': max_differences,
                     'missing_join_info': missing_join_info,
                     'hierarchical_join_info': hierarchical_join_info,
                     'dedup_info': dedup_info,
                     # can't do these filters in Jinja
                     # could write a filter function that takes a list of
                     # positive and a list of negative filter columns
                     # but this is good enough for the one case:
                     'compared_column_count': ((~column_match_info.exclude) & column_match_info.comparable & (~column_match_info.x_dtype.isnull()) & (~column_match_info.y_dtype.isnull()) & (~column_match_info.index.isin(join_cols))).sum(),
                     'column_match_info': column_match_info})


def excel_report(x_schema: str, y_schema: str,
                 x_table: str, y_table: str,
                 join_cols: list,
                 diff_summary: dict,
                 total_row_count: int,
                 column_info: dict,
                 column_match_info: pd.DataFrame,
                 missing_join_info: dict, hierarchical_join_info: dict) -> list:
    '''
    Return a list with [(sheet_name: str, df: pd.DataFrame) ... ]
    '''
    all_sheets = []
    summary_sheet_data = [{'Summary': 'Diff report between tables {x_table} (herein, "x") and {y_table} (herein, "y").'.format(
        x_table=x_table,
        y_table=y_table
    )}]
    summary_sheet_data.append({'Summary': '----'})
    summary_sheet_data.append({'Summary': 'There are {x_missing_count} rows in {y_table} that are not in {x_table}.'.format(
        x_table=x_table,
        y_table=y_table,
        x_missing_count=missing_join_info['left']['count']
    )})
    summary_sheet_data.append({'Summary': 'There are {y_missing_count} rows in {x_table} that are not in {y_table}.'.format(
        x_table=x_table,
        y_table=y_table,
        y_missing_count=missing_join_info['right']['count']
    )})
    summary_sheet_data.append({'Summary': "There are {diff_row_count} rows matched between tables that don't line up exactly.".format(diff_row_count=diff_summary['count'])})
    summary_sheet_data.append({'Summary': 'There are {column_info} columns that have differences.'.format(column_info=len(column_info))})
    if len(list(column_info.values())) > 0:
        max_differences = list(column_info.values())[0]['count']
    else:
        max_differences = 0
    summary_sheet_data.append({'Summary': 'The maximum number of differences on any individual column is {max_differences}.'.format(max_differences=max_differences)})
    all_sheets.append(('Summary', pd.DataFrame(summary_sheet_data)))

    if missing_join_info['left']['count'] > 0:
        # all_sheets.append(('Missing rows in {x_table}'.format(x_table=x_table), x_missing_ids))
        all_sheets.append(('Missing in x', missing_join_info['left']['sample']))
    if missing_join_info['right']['count'] > 0:
        # all_sheets.append(('Missing rows in {y_table}'.format(y_table=y_table), y_missing_ids))
        all_sheets.append(('Missing in y', missing_join_info['right']['sample']))
    if diff_summary['count'] > 0:
        all_sheets.append(('Mismatched rows', diff_summary['sample']))
    for column, info in column_info.items():
        all_sheets.append((column[:MAX_EXCEL_SHEET_NAME_LEN], info['df_raw']))
    return all_sheets



# This module is used by setup.py to pull the version below.
# WARNING: this approach will fail if we import anything here that
# we rely on setup.py to install.
# See that warning on Step 6 here:
# https://packaging.python.org/guides/single-sourcing-package-version/
# If we want to do imports here, there is a different approach.
__version__ = '0.5.6'



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from glob import glob
from pathlib import Path, PurePath

from setuptools import find_packages, setup

from src import dbdiff


def read(*names, **kwargs):
    with Path(PurePath.joinpath(Path(__file__).parent, *names)).open(
        encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()


setup(
    name='dbdiff',
    version=dbdiff.__version__,
    license='MIT',
    description='Compare two tables on Vertica.',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    author='Andy Reagan',
    author_email='andy@andyreagan.com',
    url='https://github.com/andyreagan/dbdiff',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[PurePath(path).name.suffix[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        # 'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        # 'Programming Language :: Python :: 3.8',
        # 'Programming Language :: Python :: Implementation :: CPython',
        # 'Programming Language :: Python :: Implementation :: PyPy',
        # uncomment if you test on these interpreters:
        # 'Programming Language :: Python :: Implementation :: IronPython',
        # 'Programming Language :: Python :: Implementation :: Jython',
        # 'Programming Language :: Python :: Implementation :: Stackless',
        'Topic :: Utilities',
    ],
    project_urls={
        # 'Documentation': 'https://dbdiff.readthedocs.io/',
        # 'Changelog': 'https://dbdiff.readthedocs.io/en/latest/changelog.html',
        'Issue Tracker': 'https://github.com/andyreagan/dbdiff/issues',
    },
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    python_requires='>=3.6',
    install_requires=[
        'click',
        'requests',
        'pandas',
        'Jinja2',
        'python-dotenv',
        'vertica_python',
        'xlsxwriter'
        # eg: 'aspectlib==1.1.1', 'six>=1.7',
    ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
    },
    entry_points={
        'console_scripts': [
            'dbdiff = dbdiff.cli:main',
        ]
    },
)



import logging
import os
from pathlib import Path

import pandas as pd
import pytest
from click.testing import CliRunner

from dbdiff.main import check_primary_key
from dbdiff.main import create_diff_table
from dbdiff.main import create_joined_table
from dbdiff.main import get_column_diffs
from dbdiff.main import get_diff_columns
from dbdiff.main import get_diff_rows
from dbdiff.main import get_unmatched_rows
from dbdiff.main import get_unmatched_rows_straight
from dbdiff.main import insert_diff_table
from dbdiff.main import select_distinct_rows
from dbdiff.cli import main
from dbdiff.vertica import get_column_info
from dbdiff.vertica import get_column_info_lookup
from dbdiff.vertica import get_cur
from dbdiff.vertica import get_table_exists

logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.INFO)
VALID_COL = {'comparable': True, 'exclude': False}
INT_DTYPES = {d: 'int' for d in {'x_dtype', 'y_dtype'}}
VARCHAR_DTYPES = {d: 'varchar(10)' for d in {'x_dtype', 'y_dtype'}}
DATE_DTYPES = {d: 'date' for d in {'x_dtype', 'y_dtype'}}
COMPARE_COLS = pd.DataFrame({'data1': {**INT_DTYPES, **VALID_COL},
                             'data2': {**INT_DTYPES, **VALID_COL},
                             'data3': {**DATE_DTYPES, **VALID_COL},
                             'data4': {**VARCHAR_DTYPES, **VALID_COL},}).transpose()
JOIN_COLS = pd.DataFrame({'join1': {**VARCHAR_DTYPES, **VALID_COL},
                          'join2': {**VARCHAR_DTYPES, **VALID_COL}}).transpose()


@pytest.fixture(scope='session')
def cur():
    os.environ['VERTICA_HOST'] = 'localhost'
    os.environ['VERTICA_PORT'] = '5433'
    os.environ['VERTICA_DATABASE'] = 'docker'
    os.environ['VERTICA_USERNAME'] = 'dbadmin'
    os.environ['VERTICA_PASSWORD'] = ''
    with get_cur() as c:
        yield c


def create_schema(cur):
    cur.execute('CREATE SCHEMA dbdiff;')


def create_x_table(cur):
    cur.execute('CREATE TABLE dbdiff.x_table ( join1 varchar(10), join2 varchar(10), missingx int, missingx2 int, dtypemiss int, data1 int, data2 int, data3 date, data4 varchar(10));')
    cur.execute("INSERT INTO dbdiff.x_table ( join1, join2, missingx, missingx2, dtypemiss, data1, data2, data3, data4) (select 'match1', 'matchdup21', 0, 0, 0, 0, 0, '2017-10-11', '');")
    cur.execute("INSERT INTO dbdiff.x_table ( join1, join2, missingx, missingx2, dtypemiss, data1, data2, data3, data4) (select 'match1', 'match22', 0, 0, 0, 0, 0, '2017-10-11', 'a');")
    cur.execute("INSERT INTO dbdiff.x_table ( join1, join2, missingx, missingx2, dtypemiss, data1, data2, data3, data4) (select 'match1', 'match23', 0, 0, 0, 1, 1, '2017-10-11', '');")
    cur.execute("INSERT INTO dbdiff.x_table ( join1, join2, missingx, missingx2, dtypemiss, data1, data2, data3, data4) (select 'match1', 'missx21', null, null, null, null, null, null, '');")
    cur.execute("INSERT INTO dbdiff.x_table ( join1, join2, missingx, missingx2, dtypemiss, data1, data2, data3, data4) (select 'match1', 'missx22', null, null, null, null, null, null, '');")
    cur.execute("INSERT INTO dbdiff.x_table ( join1, join2, missingx, missingx2, dtypemiss, data1, data2, data3, data4) (select 'missx11', null, null, null, null, null, null, null, '');")
    cur.execute("INSERT INTO dbdiff.x_table ( join1, join2, missingx, missingx2, dtypemiss, data1, data2, data3, data4) (select 'missx12', null, null, null, null, null, null, null, '');")
    cur.execute('COMMIT;')


def create_y_table(cur, case_off: bool = False):
    cur.execute('CREATE TABLE dbdiff.y_table ( join1 varchar(10), join2 varchar(10), missingy int, dtypemiss date, data1 int, data2 int, data3 date, data4 varchar(10));')
    cur.execute("INSERT INTO dbdiff.y_table ( join1, join2, missingy, dtypemiss, data1, data2, data3, data4) (select 'match1', 'matchdup21', 0, '2019-04-22', 0, 0, '2017-10-11', '');")
    cur.execute("INSERT INTO dbdiff.y_table ( join1, join2, missingy, dtypemiss, data1, data2, data3, data4) (select 'match1', 'matchdup21', 0, '2019-04-22', 0, 0, '2017-10-11', '');")
    if case_off:
        # here, we'll uppercase the 'A' so that these don't match
        cur.execute("INSERT INTO dbdiff.y_table ( join1, join2, missingy, dtypemiss, data1, data2, data3, data4) (select 'match1', 'match22', 0, '2019-04-22', 0, 1, '2017-10-12', 'A');")
    else:
        cur.execute("INSERT INTO dbdiff.y_table ( join1, join2, missingy, dtypemiss, data1, data2, data3, data4) (select 'match1', 'match22', 0, '2019-04-22', 0, 1, '2017-10-12', 'a');")
    cur.execute("INSERT INTO dbdiff.y_table ( join1, join2, missingy, dtypemiss, data1, data2, data3, data4) (select 'match1', 'match23', 0, '2019-04-22', 0, 0, '2017-10-13', '');")
    cur.execute("INSERT INTO dbdiff.y_table ( join1, join2, missingy, dtypemiss, data1, data2, data3, data4) (select 'match1', 'missy21', 0, '2019-04-22', 0, 0, null, '');")
    cur.execute("INSERT INTO dbdiff.y_table ( join1, join2, missingy, dtypemiss, data1, data2, data3, data4) (select 'missy11', null, 0, '2019-04-22', 0, 0, null, '');")
    cur.execute('COMMIT;')


def drop_schema(cur):
    cur.execute('DROP SCHEMA IF EXISTS dbdiff CASCADE;')


def test_drop_data_start(cur):
    drop_schema(cur)


def test_create_data(cur):
    create_schema(cur)
    create_x_table(cur)
    create_y_table(cur)


def test_get_column_info(cur):
    column_info = get_column_info(cur, 'dbdiff', 'x_table')
    assert type(column_info) == pd.DataFrame
    assert 'column_name' in column_info.columns
    assert 'data_type' in column_info.columns


def test_get_column_info_lookup(cur):
    column_info_lookup = get_column_info_lookup(cur, 'dbdiff', 'x_table')
    assert 'data1' in column_info_lookup
    assert column_info_lookup['join1'].lower() == 'varchar(10)'
    assert column_info_lookup['data1'] == 'int'
    assert column_info_lookup['data2'] == 'int'
    assert column_info_lookup['data3'] == 'date'
    assert len(column_info_lookup) == 9


def test_get_table_exists(cur):
    assert get_table_exists(cur, 'dbdiff', 'x_table')
    assert get_table_exists(cur, 'dbdiff', 'y_table')
    assert not get_table_exists(cur, 'dbdiff', 'z_table')


def test_check_primary_key(cur):
    assert check_primary_key(cur, 'dbdiff', 'x_table', [
                             'join1', 'join2']) == 0
    assert check_primary_key(cur, 'dbdiff', 'x_table', ['join1']) == 4


def test_select_distinct_rows(cur):
    for use_temp_tables in {True, False}:
        new_table_schema, new_table_name = select_distinct_rows(cur, 'dbdiff', 'x_table', ['join1'])
        # check that new table gets create with N rows
        assert get_table_exists(cur, new_table_schema, 'x_table_dup')
        assert get_table_exists(cur, new_table_schema, new_table_name)
        cur.execute('select * from {schema}.{table}'.format(schema=new_table_schema, table=new_table_name))
        df = pd.DataFrame(cur.fetchall())
        assert df.shape[0] == 2
        assert df.shape[1] == 9
        cur.execute('select * from {schema}.{table}'.format(schema=new_table_schema, table='x_table_dup'))
        df = pd.DataFrame(cur.fetchall())
        assert df.shape[0] == 5
        assert df.shape[1] == 10


def test_create_joined_table(cur):
    create_joined_table(cur, x_schema='dbdiff', y_schema='dbdiff',
                        x_table='x_table', y_table='y_table',
                        join_cols=['join1', 'join2'],
                        compare_cols=pd.concat([COMPARE_COLS, JOIN_COLS]),
                        joined_schema='dbdiff', joined_table='x_table_JOINED')
    # check that it is created and has the right number of rows, columns...
    assert get_table_exists(cur, 'dbdiff', 'x_table_JOINED')
    cur.execute('select * from {schema}.{table}'.format(schema='dbdiff', table='x_table_JOINED'))
    df = pd.DataFrame(cur.fetchall())
    assert df.shape[0] == 4
    assert df.shape[1] == 10


def test_get_unmatched_rows_straight(cur):
    join_cols = ['join1', 'join2']
    missing_left_join = [1, 1]
    missing_right_join = [2, 2]
    missing_left_total = sum(missing_left_join)
    missing_right_total = sum(missing_right_join)
    results = get_unmatched_rows_straight(cur, 'dbdiff', 'dbdiff', 'x_table', 'y_table',
                                          join_cols, 100)
    expected_results = {'left': {'count': missing_left_total,
                                 'sample_shape': (missing_left_total, len(join_cols))},
                        'right': {'count': missing_right_total,
                                  'sample_shape': (missing_right_total, len(join_cols))}}
    assert results['left']['count'] == expected_results['left']['count']
    assert results['left']['sample'].shape[0] == expected_results['left']['sample_shape'][0]
    assert results['left']['sample'].shape[1] == expected_results['left']['sample_shape'][1]
    assert results['right']['count'] == expected_results['right']['count']
    assert results['right']['sample'].shape[0] == expected_results['right']['sample_shape'][0]
    assert results['right']['sample'].shape[1] == expected_results['right']['sample_shape'][1]


def test_get_unmatched_rows(cur):
    join_cols = ['join1', 'join2']
    missing_left_join = [1, 1]
    missing_right_join = [2, 2]
    results = get_unmatched_rows(cur, 'dbdiff', 'dbdiff', 'x_table', 'y_table',
                                 join_cols, 100)
    expected_results = {j: {side: {'count': d[i], 'sample_shape': (d[i], i + 1)} for side, d in (('left', missing_left_join), ('right', missing_right_join))} for i, j in enumerate(join_cols)}
    for col, expected in expected_results.items():
        logging.info(col)
        for side, expected_info in expected.items():
            logging.info(side)
            logging.info(results[col][side]['count'])
            logging.info(results[col][side]['sample'])
            assert 'sample' in results[col][side]
            assert 'query' in results[col][side]
            assert results[col][side]['count'] == expected_info['count']
            for i in {0, 1}:
                assert results[col][side]['sample'].shape[i] == expected_info['sample_shape'][i]
            if col == 'join2':
                assert 'sample_grouped' in results[col][side]
                assert 'query_grouped' in results[col][side]


def test_create_diff_table(cur):
    create_diff_table(cur, 'dbdiff', 'x_table_DIFF', ['join1', 'join2'],
                      pd.concat([COMPARE_COLS, JOIN_COLS]))
    assert get_table_exists(cur, 'dbdiff', 'x_table_DIFF')


def test_insert_diff_table(cur):
    cur.execute('select * from dbdiff.x_table_JOINED')
    logging.info(cur.fetchall())
    cur.execute('select * from dbdiff.x_table_DIFF')
    logging.info(cur.fetchall())
    insert_diff_table(cur, joined_schema='dbdiff', joined_table='x_table_JOINED',
                      diff_schema='dbdiff', diff_table='x_table_DIFF', join_cols=['join1', 'join2'], column='data1')
    cur.execute('select * from {schema}.{table}'.format(schema='dbdiff', table='x_table_DIFF'))
    df = pd.DataFrame(cur.fetchall())
    assert df.shape[0] == 1
    assert df.shape[1] == 3
    insert_diff_table(cur, joined_schema='dbdiff', joined_table='x_table_JOINED',
                      diff_schema='dbdiff', diff_table='x_table_DIFF', join_cols=['join1', 'join2'], column='data2')
    cur.execute('select * from {schema}.{table}'.format(schema='dbdiff', table='x_table_DIFF'))
    df = pd.DataFrame(cur.fetchall())
    assert df.shape[0] == 3
    assert df.shape[1] == 3


# def test_implicit_dytpe_comparison():
#     implicit_dytpe_comparison(x_dtype, y_dtype)


def test_get_diff_rows(cur):
    diff_summary = get_diff_rows(
        cur, 'dbdiff', 'x_table', ['join1', 'join2'], 100)
    assert diff_summary['count'] == 2
    assert diff_summary['total_count'] == 3
    assert diff_summary['sample'].shape[0] == 2
    # assert diff_summary['sample'].shape[1] == 10


def test_get_diff_columns(cur):
    df = get_diff_columns(cur, 'dbdiff', 'x_table')
    assert df.shape[0] == 2
    assert df.shape[1] == 2


def test_get_column_diffs(cur):
    join_cols = ['join1', 'join2']
    diff_columns = get_diff_columns(cur, 'dbdiff', 'x_table')

    grouped_column_diffs = get_column_diffs(
        diff_columns, cur, 'dbdiff', 'dbdiff', 'x_table',
        'dbdiff', 'y_table', ['join1', 'join2'], 100, COMPARE_COLS, True)
    logging.info(grouped_column_diffs)

    data1_misses = 1
    data2_misses = 2

    expected = {'data1': {'count': data1_misses, 'df_shape': (data1_misses, 3),
                          'df_raw_shape': (data1_misses, 2 + len(join_cols)),
                          'df_h_x_shape': (5, 1 + len(join_cols)),
                          'df_h_y_shape': (5, 1 + len(join_cols))},
                'data2': {'count': data2_misses, 'df_shape': (data2_misses, 3),
                          'df_raw_shape': (data2_misses, 2 + len(join_cols)),
                          'df_h_x_shape': (5, 1 + len(join_cols)),
                          'df_h_y_shape': (5, 1 + len(join_cols))}}

    for column_name in expected.keys():
        grouped_column_diffs[column_name]
        logging.info(grouped_column_diffs[column_name])
        assert expected[column_name]['count'] == grouped_column_diffs[column_name]['count']
        for q_name in {'q', 'q_raw', 'q_h_x', 'q_h_y'}:
            assert q_name in grouped_column_diffs[column_name]
        for i in {0, 1}:
            assert expected[column_name]['df_shape'][i] == grouped_column_diffs[column_name]['df'].shape[i]
            assert expected[column_name]['df_raw_shape'][i] == grouped_column_diffs[column_name]['df_raw'].shape[i]
            assert expected[column_name]['df_h_x_shape'][i] == grouped_column_diffs[column_name]['df_h_x'].shape[i]
            assert expected[column_name]['df_h_y_shape'][i] == grouped_column_diffs[column_name]['df_h_y'].shape[i]


def test_drop_data_end(cur):
    drop_schema(cur)


def test_main(cur):
    create_schema(cur)
    create_x_table(cur)
    create_y_table(cur)

    base_options = ['dbdiff', 'x_table', 'y_table', 'join1,join2']
    runner = CliRunner()

    def runner_wrapper(runner, base_options, addl_options):
        result = runner.invoke(
            main, base_options + addl_options)
        if result.exit_code:
            logging.info(str(result.exception) + str(result.exc_info))
        assert result.exit_code == 0

    runner_wrapper(runner, base_options, [])
    # save the report:
    # Path('x_table_report.html').rename('base_report.html')
    # clear the output
    drop_schema(cur)
    create_schema(cur)
    create_x_table(cur)
    create_y_table(cur)

    runner_wrapper(runner, base_options, ['--drop-output-tables'])
    runner_wrapper(runner, base_options, ['--drop-output-tables', '--output-format=XLSX'])
    runner_wrapper(runner, base_options, ['--hierarchical-join'])
    runner_wrapper(runner, base_options, ['--drop-output-tables'])
    runner_wrapper(runner, base_options, ['--drop-output-tables', '--output-format=XLSX'])
    runner_wrapper(runner, base_options, ['--use-diff-table'])
    runner_wrapper(runner, base_options, ['--hierarchical-join', '--use-diff-table'])

    Path('x_table_report.html').unlink()
    Path('x_table_report.xlsx').unlink()

    Path('x_table_temp.sql').write_text('select * from dbdiff.x_table')
    x_table_temp_options = ['dbdiff', 'x_table_temp.sql', 'y_table', 'join1,join2', '--x-table-query']
    runner_wrapper(runner, x_table_temp_options, ['--drop-output-tables'])
    runner_wrapper(runner, x_table_temp_options, ['--drop-output-tables', '--output-format=XLSX'])
    runner_wrapper(runner, x_table_temp_options, ['--hierarchical-join'])

    Path('y_table_temp.sql').write_text('select * from dbdiff.y_table')
    both_table_temp_options = ['dbdiff', 'x_table_temp.sql', 'y_table_temp.sql', 'join1,join2', '--x-table-query', '--y-table-query']
    runner_wrapper(runner, both_table_temp_options, ['--drop-output-tables'])
    runner_wrapper(runner, both_table_temp_options, ['--drop-output-tables', '--output-format=XLSX'])
    runner_wrapper(runner, both_table_temp_options, ['--hierarchical-join'])
    runner_wrapper(runner, both_table_temp_options, ['--save-column-summary'])
    runner_wrapper(runner, both_table_temp_options, ['--save-column-summary', '--save-column-summary-format=pickle'])


    Path('x_table_temp.sql').unlink()
    Path('y_table_temp.sql').unlink()
    Path('x_table_temp_report.html').unlink()
    Path('x_table_temp_report.xlsx').unlink()
    Path('x_table_temp_col_info.csv').unlink()
    Path('x_table_temp_col_info.pkl').unlink()

    # clear the output
    drop_schema(cur)
    create_schema(cur)
    create_x_table(cur)
    create_y_table(cur, case_off=True)

    runner_wrapper(runner, base_options, [])
    # Path('x_table_report.html').rename('case_on.html')
    runner_wrapper(runner, base_options, ['--case-insensitive'])
    # Path('x_table_report.html').rename('case_off.html')

    Path('x_table_report.html').unlink()